unit poly_unit;  //polygon procedures and functions

interface

uses types,graphics,sysutils;

const maxpoint = 40;
      antivirus = 'zomaar wat flauwekul';

type TFpoint = record
                x : single;
                y : single;
               end;
     TFpoints = array[1..maxpoint] of TFpoint;

var polyresultcode : byte;
    polyTime : longInt;
    Fpoints : TFpoints;
    pcount : byte;      //number of points

procedure setcanvas(cvs : Tcanvas);//select canvas for draw/fill of polygon
procedure polyExpand(EF : boolean);

implementation

uses unit1,clock_unit;

const offset = 0.025;           //displacement
      pi05 = 0.5*pi;
      pi15 = 1.5*pi;
      pi2 = 2*pi;
      rd05 = 0.5;                  //rounding
      frnd = 1e-6;                 //floating point round to zero
      resultmessage : array[0..5] of string =
                      ('OK',
                       'insufficient points',
                       'polygon not closed',
                       'wrong polygon',
                       'duplicate point',
                       'intersecting edges');

type
     Tvector = record
                x,y   : single;
                dx,dy : single;
                dir   : single;       //direction 0..+-pi
                modulus : single;
               end;

     TVectorList = array[1..maxpoint] of TVector;

var vectorlist : TVectorlist;
    vcount : byte;
    fvalid : boolean = false;
    f1,f2 : single;
    cv : Tcanvas = nil;

// ---- low level support ----

function VDir(deltaX,deltaY : single) : single;
//return direction of vector in radians
//(+,0) = 0; (0,+) = 0.5pi ; (-,0) = pi ; (0,-) = 1.5pi
begin
 if deltaX = 0 then
  begin
   if deltaY > 0 then result := pi05 else result := pi15;
   exit;
  end;
 result := arctan((deltaY)/(deltaX));
 if deltaX < 0 then result := result + pi;
 if result < 0 then result := result + pi2;
end;

function V12angle(dir1,dir2 : single) : single;
//return angle between vectors v1,v2 in radians
begin
 result := dir2 - dir1;
 if result > pi then result := result-pi2
  else if result < -pi then result := result+pi2;
end;

procedure vrsect(const v1,v2 : TVector);
//calculate intersection of vectors v1,v2
//line1 = (v1.x1,v1.y1) +f1*(v1.dx,v1.dy)
//line2 = (v2.x1,v2.y1) +f2*(v2.dx,v2.dy)
//return f1,f2,fvalid
var d,vx,vy : single;
begin
 d := v1.dx*v2.dy - v1.dy*v2.dx;//discriminant
 if d = 0 then begin
                fvalid := false; exit;
               end;
 fvalid := true;
 vx := v2.x - v1.x;
 vy := v2.y - v1.y;
 f1 := (vx*v2.dy - vy*v2.dx)/d;
 f2 := (vx*v1.dy - vy*v1.dx)/d;

 if abs(f1) < frnd then f1 := 0;  //round to 1e-6
 if abs(f2) < frnd then f2 := 0;
 if abs(f1-1) < frnd then f1 := 1;
 if abs(f2-1) < frnd then f2 := 1;
end;

function Adjacent(a,b : word): boolean;
//called by "checkvectors"
//return true if vectors Vlist[a],Vlist[b] share start- endpoint
begin
 result := ((a = 1) and (b = vcount)) or
           ((b = 1) and (a = vcount)) or
           (abs(a-b) < 2);
end;

//---- hi level support procedures ----

procedure buildVectorlist;
var n : word;
begin
 for n := 1 to pcount-1 do
  with vectorlist[n] do
  begin
   x := fpoints[n].x;
   y := fpoints[n].y;
   dx := fpoints[n+1].x - x;
   dy := fpoints[n+1].y - y;
   dir := VDir(dx,dy);       //direction of vector
   modulus := sqrt(sqr(dx)+sqr(dy));
  end;//with
 vcount := pcount-1; 
end;

//--- check points ----

procedure checkpoints;
//check closed polygon and sufficient points
begin
 if pcount < 3 then
  begin
   polyresultcode := 1;
   exit;
  end;

 if (fpoints[1].x <> fpoints[pcount].x) or
    (fpoints[1].y <> fpoints[pcount].y) then
  begin
   polyresultcode := 2;
   exit;
  end;
end;

function SumAngles : single;
var i : byte;
begin
 result := 0;
 for i := 1 to vcount-1 do
   result := result + V12Angle(vectorlist[i].dir,vectorlist[i+1].dir);
 result := result + V12Angle(vectorlist[vcount].dir,vectorlist[1].dir);
end;

procedure checkvectors;
//check Vectors for
// - duplicate points
// - intersections
var i,j : word;
begin
 for i := 1 to pcount-2 do                    //check duplicate points
 for j := i+1 to pcount-1 do
  if (vectorlist[i].x = vectorlist[j].x) and
     (vectorlist[i].y = vectorlist[j].y) then
   begin
    polyResultCode := 4;             //duplicate point
    exit;
   end;

 for i := 1 to vcount-1 do        //check for no intersection
  for j := i+1 to vcount do
   if adjacent(i,j) = false then
    begin
     vrsect(vectorlist[i],vectorlist[j]);
     if fvalid and ((f1 >= 0) and (f1 <= 1)) and ((f2 >= 0) and (f2 <= 1)) then
      begin
       polyResultcode := 5;  //intersection of edges
       exit;
      end;
    end;
end;

procedure shiftvector(var v : Tvector; R : boolean);
//R : true for right shift
var dd,m: single;
begin
 if R then m := 1 else m := -1;
 with v do
  begin
   dd := offset/modulus;
   x := x+dd*dy*m;
   y := y-dd*dx*m;
  end;
end;

//--- main calls ---

procedure setcanvas(cvs : Tcanvas);
begin
 cv := cvs;
end;

procedure polyExpand(EF : boolean);
//ef : expansion flag
var t1,t2 : Int64;
    i,n : byte;
    anglesum : single;
    s : string;
    CCW : boolean;
    RF : boolean;   //true if right shift of vector
label exitlabel;
begin
 t1 := getCPUticks;
 polyresultcode := 0;
 checkpoints;
 if polyresultcode <> 0 then goto exitlabel;
//
 buildvectorlist;
 checkvectors;
 if polyresultcode <> 0 then goto exitlabel;
//
 AngleSum := round(SumAngles*180/pi);
 if abs(AngleSum) <> 360 then
  begin
   polyresultcode := 3;
   goto exitLabel;
  end;
 CCW := AngleSum > 0;
 RF := not (EF xor CCW);
 for i := 1 to vcount do shiftvector(vectorlist[i],RF);
 for i := 1 to vcount do
  begin
   if i = 1 then n := vcount else n := i-1;
   vrsect(vectorlist[i],vectorlist[n]);
   with vectorlist[i] do
    begin
     Fpoints[i].x := x + f1*dx;
     Fpoints[i].y := y + f1*dy;
    end;
  end;//for
 Fpoints[pcount] := Fpoints[1];

exitlabel:

 t2 := getCPUticks;
 s := resultmessage[polyresultcode];
 if polyresultcode = 0 then
  begin
   polyTime := ProcTime(t2-t1);
   s := s + '  time= '+formatfloat('##0.###',polytime)+' usecs';
   repaintall;
  end;
 form1.msgtext.Caption := s;
end;

end.
